/* I2S es8311 Example

    This example code is in the Public Domain (or CC0 licensed, at your option.)

    Unless required by applicable law or agreed to in writing, this
    software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
    CONDITIONS OF ANY KIND, either express or implied.
*/

#include <stdio.h>
#include <string.h>
#include <sys/unistd.h>
#include <sys/stat.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "driver/i2s.h"
#include "driver/i2c.h"
#include "esp_system.h"
#include "esp_log.h"
#include "esp_vfs_fat.h"
#include "sdmmc_cmd.h"
#include "driver/sdmmc_defs.h"
#include "driver/sdmmc_host.h"
#include "es8311.h"

/**
 * Example Target: ESP32-C3 / ESP32-S3
 * This example is a referenct example,
 * mainly aims to offer a way of transporting i2s read data to another thread
 *
 * NOTICE: When the sample rate is high, there will be a large quantity of data received
 * In this case, we need a large buffer to store these data (refer EXAMPLE_RECV_BUF_SIZE in this example)
 * If the buffer is not big enough, the period of every read cycle would be shorter,
 * that means `i2s_read` will spend less time on receiving the data from ISR
 * Besides, you would have less time on processing data in another thread as well.
 * And data lost will take place when the buffer is ran out.
 *
 * HOW TO ESTIMATE THE RECEIVE BUFFER SIZE
 * As mentioned above, we need a big enough buffer while transporting data.
 * In this example, we use 'recv_msg_t' to carry the data buffer and other infomations
 * Firstly we should know how long it will take to processing 'EXAMPLE_RECV_BUF_SIZE' length data
 * For example, it takes 5ms， mark it as symbol 'Tp'
 * Then we can also get the 'i2s_read' block time in this example, marked as 'Tr'
 * ('Tr' can be calculated as well, please refer to 'overtime_tick')
 * We need to guarantee that 'Tr' > 'Tp', otherwise processing speed can't catch the reading speed, data will be dropped
 * If 'Tr' <= 'Tp', try to increase 'EXAMPLE_RECV_BUF_SIZE' until 'Tr' > 'Tp'
 * For addition, normally two 'recv_msg_t' is enough, we only need to increase 'EXAMPLE_RECV_BUF_SIZE' to meet 'Tr' > 'Tp' condition
 */

/* I2C port and GPIOs */
#define I2C_NUM         (0)
#define I2C_SDA_IO      (GPIO_NUM_5)
#define I2C_SCL_IO      (GPIO_NUM_6)

/* I2S port and GPIOs */
#define I2S_NUM         (0)
#define I2S_MCK_IO      (GPIO_NUM_3)
#define I2S_BCK_IO      (GPIO_NUM_7)
#define I2S_WS_IO       (GPIO_NUM_16)
#define I2S_DO_IO       (GPIO_NUM_17)
#define I2S_DI_IO       (GPIO_NUM_18)

/* I2S configurations */
#define EXAMPLE_DMA_FRAME_NUM   (200)
#define EXAMPLE_DMA_DESC_NUM    (8)
#define EXAMPLE_I2S_CHAN_NUM    (2)
#define EXAMPLE_SAMPLE_RATE     (48000)
#define EXAMPLE_MCLK_MULTIPLE   I2S_MCLK_MULTIPLE_256
#define EXAMPLE_SAMPLE_BIT      (16)

/* ES8311 configurations */
#define EXAMPLE_VOICE_VOLUME    (60)
#define EXAMPLE_MIC_GAIN        ES8311_MIC_GAIN_18DB

/* Receive task configurations */
#define EXAMPLE_RECV_BUF_SIZE   (4800)
#define EXAMPLE_MSG_QUEUE_NUM   (2)
#define EXAMPLE_CPU_FREQ        CONFIG_ESP32S3_DEFAULT_CPU_FREQ_MHZ

/* SD card & recording task configurations */
#define RECORDING_TIME_SEC      (5)
#define FLASH_RECORD_SIZE       (RECORDING_TIME_SEC * EXAMPLE_SAMPLE_RATE * EXAMPLE_I2S_CHAN_NUM * EXAMPLE_SAMPLE_BIT / 8)
#define SD_MOUNT_POINT          "/sdcard"
#define SPI_DMA_CHAN            (2)

typedef struct recv_msg_t recv_msg_t;

struct recv_msg_t{
    recv_msg_t  *self;
    SemaphoreHandle_t mutex;
    uint32_t    len;
    uint8_t     *buf;
};


static const char *TAG = "i2s_record";
static bool record_finish = false;

// When testing SD and SPI modes, keep in mind that once the card has been
// initialized in SPI mode, it can not be reinitialized in SD mode without
// toggling power to the card.
sdmmc_host_t host = SDMMC_HOST_DEFAULT();
QueueHandle_t msg_que;
QueueHandle_t evt_que;
sdmmc_card_t *card;


static void es8311_init(void)
{
    i2c_config_t es_i2c_cfg = {
        .sda_io_num = I2C_SDA_IO,
        .scl_io_num = I2C_SCL_IO,
        .mode = I2C_MODE_MASTER,
        .sda_pullup_en = GPIO_PULLUP_ENABLE,
        .scl_pullup_en = GPIO_PULLUP_ENABLE,
        .master.clk_speed = 100000,
    };
    audio_hal_codec_config_t es8311_cfg = {
        .adc_input  = AUDIO_HAL_ADC_INPUT_LINE1,
        .dac_output = AUDIO_HAL_DAC_OUTPUT_ALL,
        .codec_mode = AUDIO_HAL_CODEC_MODE_BOTH,
        .i2s_iface = {
            .mode = AUDIO_HAL_MODE_SLAVE,
            .fmt = AUDIO_HAL_I2S_NORMAL,
            .samples = AUDIO_HAL_48K_SAMPLES,
            .bits = AUDIO_HAL_BIT_LENGTH_16BITS,
        },
    };
    es8311_codec_init(&es8311_cfg, I2C_NUM, &es_i2c_cfg);
    es8311_codec_config_i2s(es8311_cfg.codec_mode, &es8311_cfg.i2s_iface);
    es8311_set_mic_gain(EXAMPLE_MIC_GAIN);
    es8311_codec_set_voice_volume(EXAMPLE_VOICE_VOLUME);
    es8311_codec_ctrl_state(AUDIO_HAL_CODEC_MODE_BOTH, AUDIO_HAL_CTRL_START);
}

static void i2s_driver_init(void)
{
    i2s_config_t i2s_cfg = {
        .mode = I2S_MODE_MASTER | I2S_MODE_RX,
        .sample_rate = EXAMPLE_SAMPLE_RATE,
        .bits_per_sample = EXAMPLE_SAMPLE_BIT,
        .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
        .communication_format = I2S_COMM_FORMAT_STAND_I2S,
        .tx_desc_auto_clear = false,
#if SOC_I2S_SUPPORTS_TDM
        .total_chan = EXAMPLE_I2S_CHAN_NUM,
        .chan_mask = I2S_TDM_ACTIVE_CH0 | I2S_TDM_ACTIVE_CH1,
        .left_align = false,
        .big_edin = false,
        .bit_order_msb = false,
        .skip_msk = false,
#endif
        .dma_buf_count = EXAMPLE_DMA_DESC_NUM,
        .dma_buf_len = EXAMPLE_DMA_FRAME_NUM,
        .use_apll = false,
        .mclk_multiple = EXAMPLE_MCLK_MULTIPLE,
        .intr_alloc_flags = ESP_INTR_FLAG_LEVEL1,
    };

    i2s_driver_install(I2S_NUM, &i2s_cfg, EXAMPLE_MSG_QUEUE_NUM, &evt_que);
    i2s_pin_config_t i2s_pin_cfg = {
        .mck_io_num = I2S_MCK_IO,
        .bck_io_num = I2S_BCK_IO,
        .ws_io_num = I2S_WS_IO,
        .data_out_num = I2S_DO_IO,
        .data_in_num = I2S_DI_IO
    };
    i2s_set_pin(I2S_NUM, &i2s_pin_cfg);
}

static void mount_sdcard(const char *basePath)
{
    host.flags = SDMMC_HOST_FLAG_1BIT;

    sdmmc_slot_config_t slot_config = SDMMC_SLOT_CONFIG_DEFAULT();
    // slot_config.gpio_cd = SD_CARD_INTR_GPIO;
    slot_config.width = 4;

    esp_vfs_fat_sdmmc_mount_config_t mount_config = {
        .format_if_mount_failed = true,
        .max_files = 5,
        .allocation_unit_size = 16 * 1024
    };


    esp_err_t ret;
retry:
    ret = esp_vfs_fat_sdmmc_mount(basePath, &host, &slot_config, &mount_config, &card);

    if (ret != ESP_OK) {
        if (ret == ESP_FAIL) {
            ESP_LOGE(TAG, "Failed to mount filesystem. "
                     "If you want the card to be formatted, set the EXAMPLE_FORMAT_IF_MOUNT_FAILED menuconfig option.");
        } else {
            ESP_LOGE(TAG, "Failed to initialize the card (%s). "
                     "Make sure SD card lines have pull-up resistors in place.", esp_err_to_name(ret));
        }
        ESP_LOGI(TAG, "retry in 3 sec");
        vTaskDelay(300);
        goto retry;
    }
    sdmmc_card_print_info(stdout, card);
}

static void sd_record_task(void *args)
{
    int flash_wr_size = 0;
    recv_msg_t msg;
    const char filename[] = SD_MOUNT_POINT"/record";

retry:
    ESP_LOGI(TAG, "Opening new file");
    // First check if file exists before creating a new file.
    struct stat st;
    printf("Filename: %s\n", filename);
    if (stat(filename, &st) == 0) {
        // Delete it if it exists
        unlink(filename);
    }

    // Create a new file
    FILE *f = fopen(filename, "w");
    if (f == NULL) {
        ESP_LOGE(TAG, "Failed to open file for writing");
        vTaskDelay(300);
        goto retry;
    }

    while (flash_wr_size < FLASH_RECORD_SIZE) {
        // receive from evnet queue
        xQueueReceive(msg_que, &msg, portMAX_DELAY);
        // Take the mutex to avoid the buffer overwriten by reading task
        xSemaphoreTake(msg.self->mutex, portMAX_DELAY);
        // Write raw sample data to the file
        fwrite(msg.buf, 1, msg.len, f);
        xSemaphoreGive(msg.self->mutex);
        flash_wr_size += msg.len;
    }
    record_finish = true;
    ESP_LOGI(TAG, "Recording done!");
    fclose(f);
    esp_vfs_fat_sdcard_unmount(SD_MOUNT_POINT, card);
    ESP_LOGI(TAG, "File written on SDCard");
    while (1) {
        vTaskDelay(5000 / portTICK_RATE_MS);
    }
    vTaskDelete(NULL);
}

static void i2s_read_task(void *args)
{
    static recv_msg_t msg[EXAMPLE_MSG_QUEUE_NUM];
    for (int i = 0; i < EXAMPLE_MSG_QUEUE_NUM; i++) {
        msg[i].self = &msg[i];
        msg[i].mutex = xSemaphoreCreateMutex();
        msg[i].len = 0;
        msg[i].buf = calloc(EXAMPLE_RECV_BUF_SIZE, sizeof(uint8_t));
        if (!msg[i].buf) {
            ESP_LOGE(TAG, "No memory for read data buffer");
            abort();
        }
    }
    // If waiting for mutex more than 'overtime_tick' it means there might be data dropped
    uint32_t overtime_tick = ((EXAMPLE_RECV_BUF_SIZE * 1000 * 8) / (EXAMPLE_SAMPLE_RATE * EXAMPLE_I2S_CHAN_NUM * EXAMPLE_SAMPLE_BIT)) / portTICK_RATE_MS;
    uint32_t t_start = 0, t_end = 0;

    for (int cnt = 0;; cnt++, cnt %= EXAMPLE_MSG_QUEUE_NUM) {
        // Waiting utill the buffer is released by recording task
        // Set overtime to 'overtime_tick' because if reading threading waiting for more than 'overtime_tick', there might be data lost
        if (xSemaphoreTake(msg[cnt].mutex, overtime_tick) == pdTRUE) {
            /* Read sample data from mic */
            t_start = esp_cpu_get_ccount();
            i2s_read(I2S_NUM, msg[cnt].buf, EXAMPLE_RECV_BUF_SIZE, &msg[cnt].len, portMAX_DELAY);
            t_end = esp_cpu_get_ccount();
            printf("i2s_read block time: %d us\n", (t_end - t_start) / EXAMPLE_CPU_FREQ);
            xSemaphoreGive(msg[cnt].mutex);
            // Set overtime portMAX_DELAY here to block reading thread once recording finished
            // You can set a properiate overtime value for your application
            xQueueSend(msg_que, (void *) &msg[cnt], portMAX_DELAY);
        } else {
            ESP_LOGW(TAG,"Waiting for recording task overtime");
        }
    }
    vTaskDelete(NULL);
}

static void i2s_event(void *args)
{
    i2s_event_t evt;
    uint32_t cnt = 0;
    while (!record_finish) {
        xQueueReceive(evt_que, &evt, portMAX_DELAY);
        if (evt.type == I2S_EVENT_RX_Q_OVF) {
            ESP_LOGW(TAG,"[%d] i2s queue overflow, data dropped", cnt++);
        }
    }
    vTaskDelete(NULL);
}

void app_main(void)
{
    mount_sdcard(SD_MOUNT_POINT);
    /* Initialize i2s peripheral */
    i2s_driver_init();
    /* Initialize i2c peripheral and config es8311 codec by i2c */
    es8311_init();
    /* message queue */
    msg_que = xQueueCreate(EXAMPLE_MSG_QUEUE_NUM, sizeof(recv_msg_t));
    /* Record sample data in SD card */
    xTaskCreatePinnedToCore(sd_record_task, "sd_record_task", 4096, NULL, 6, NULL, 1);
    /* I2S reading task */
    xTaskCreatePinnedToCore(i2s_read_task, "i2s_read_task", 8192, NULL, 7, NULL, 0);
    /* Receiving I2S recv event, show warning if data dropped */
    xTaskCreatePinnedToCore(i2s_event, "i2s_event", 4096, NULL, 5, NULL, 0);
}
